import asyncio
import time

import discord
from typing_extensions import Self
from v6d2ctx.integration.responsetype import *
from v6d2ctx.integration.targets import *

from rainbowadn.instrument import Instrumentation
from v6d2ctx.context import *
from v6d3music.core.mainaudio import *
from v6d3music.core.mainservice import *

__all__ = ('Api',)


class Api:
    class MisusedApi(Exception):
        def json(self) -> dict:
            return {'error': list(map(str, self.args)), 'errormessage': str(self)}

    class UnknownApi(MisusedApi):
        def json(self) -> dict:
            return super().json() | {'unknownapi': None}

    class ExplicitFailure(MisusedApi):
        def __init__(self, explicit: Explicit) -> None:
            super().__init__(*explicit.args)

        def json(self) -> dict:
            return super().json() | {'explicit': None}

    def __init__(self, mainservice: MainService, roles: dict[str, str]) -> None:
        self.mainservice = mainservice
        self.client = mainservice.client
        self.roles = roles
        self.targets = mainservice.targets
        self.targets.register_instance(self, 'api', Async)
        self.targets.register_instrumentation('Count', lambda t, n: Count(t, n))
        self.targets.register_instrumentation('Concurrency', lambda t, n: Concurrency(t, n), Async)

    def user_id(self) -> int | None:
        if self.client.user is None:
            return None
        else:
            return self.client.user.id

    def is_operator(self, user_id: int) -> bool:
        return '(operator)' in self.roles.get(f'roles{user_id}', '')

    async def api(self, request: dict, user_id: int) -> ResponseType:
        response = await UserApi(ApiSession(self), request, user_id).api()
        match response, request:
            case {'time': _}, _:
                pass
            case dict() as d, {'time': _}:
                response = d | {'time': time.time()}
        return response


class ApiSession:
    def __init__(self, api: Api) -> None:
        self.__api = api
        self.__complexity = 1000

    def api(self):
        if self.__complexity <= 0:
            raise Api.MisusedApi('hit complexity limit')
        self.__complexity -= 1
        return self.__api


class UserApi:
    class UnknownMember(Api.MisusedApi):
        def json(self) -> dict:
            return super().json() | {'unknownmember': None}

    def __init__(self, session: ApiSession, request: dict, user_id: int) -> None:
        self.session = session
        self.pi = session.api()
        self.client = self.pi.client
        self.request = request
        self.user_id = user_id
        self._parent: Self | None = None
        self._key: int | str | None = None

    async def subs(self, requests: list[dict] | dict[str, dict]) -> ResponseType:
        match self.request:
            case {'idkey': str() as idkey}:
                pass
            case _:
                idkey = 'type'
        match self.request:
            case {'idbase': dict() as base}:
                pass
            case _:
                base = {}
        match requests:
            case list():
                return list(
                    await asyncio.gather(
                        *(self.sub(request, key).api() for (key, request) in enumerate(requests))
                    )
                )
            case dict():
                items = list(requests.items())
                responses = await asyncio.gather(
                    *(self.sub({idkey: key} | base | request, key).api() for key, request in items)
                )
                return dict((key, response) for (key, _), response in zip(items, responses))
            case _:
                raise Api.MisusedApi('that should not happen')

    def _sub(self, request: dict) -> Self:
        return UserApi(self.session, request, self.user_id)

    def sub(self, request: dict, key: str | int) -> Self:
        sub = self._sub(request)
        sub._parent = self
        sub._key = key
        return sub

    async def to_guild_api(self, guild_id: int) -> 'GuildApi':
        guild = self.client.get_guild(guild_id) or await self.client.fetch_guild(guild_id)
        if guild is None:
            raise UserApi.UnknownMember('unknown guild')
        member = guild.get_member(self.user_id) or await guild.fetch_member(self.user_id)
        if member is None:
            raise UserApi.UnknownMember('unknown member of a guild')
        return GuildApi(self, member)

    async def to_operator_api(self) -> 'OperatorApi':
        if not self.pi.is_operator(self.user_id):
            raise UserApi.UnknownMember('not an operator')
        return OperatorApi(self)

    def _api_text(self) -> str:
        return 'user api'

    async def _fall_through_api(self) -> ResponseType:
        match self.request:
            case {'type': '?'}:
                return f'this is {self._api_text()}'
            case {'type': '*', 'requests': list() | dict() as requests}:
                return await self.subs(requests)
            case _:
                raise Api.UnknownApi(f'unknown {self._api_text()}')

    async def _api(self) -> ResponseType:
        match self.request:
            case {'guild': str() as guild_id_str} if guild_id_str.isdecimal() and len(guild_id_str) < 100:
                self.request.pop('guild')
                return await (await self.to_guild_api(int(guild_id_str))).api()
            case {'operator': _}:
                self.request.pop('operator')
                return await (await self.to_operator_api()).api()
            case {'type': 'ping', 't': (float() | int()) as t}:
                return time.time() - t
            case {'type': 'guilds'}:
                guilds = []
                for guild in self.client.guilds:
                    if guild.get_member(self.user_id) is not None:
                        guilds.append(str(guild.id))
                return guilds
            case _:
                return await self._fall_through_api()

    async def api(self) -> ResponseType:
        try:
            try:
                return await self._api()
            except Explicit as e:
                raise Api.ExplicitFailure(e) from e
        except Api.MisusedApi as e:
            catches = self.request.get('catches', {})
            if len(e.args) and (key := e.args[0]) in catches:
                return catches[key]
            if '*' in catches:
                return e.json()
            raise


class GuildApi(UserApi):
    class VoiceNotConnected(Api.MisusedApi):
        def json(self) -> dict:
            return super().json() | {'notconnected': None}

    def __init__(self, api: UserApi, member: discord.Member) -> None:
        super().__init__(api.session, api.request, member.id)
        self.member = member
        self.guild = member.guild

    async def to_voice_api(self) -> 'VoiceApi':
        voice = self.member.voice
        if voice is None:
            raise GuildApi.VoiceNotConnected('you are not connected to voice')
        channel = voice.channel
        if channel is None:
            raise GuildApi.VoiceNotConnected('you are not connected to a voice channel')
        if self.client.user is None:
            raise GuildApi.VoiceNotConnected('bot client user not initialised')
        if self.client.user.id not in channel.voice_states:
            raise GuildApi.VoiceNotConnected('bot not connected')
        return VoiceApi(self, channel)

    def _sub(self, request: dict) -> Self:
        return GuildApi(super()._sub(request), self.member)

    def _api_text(self) -> str:
        return 'guild api'

    async def _api(self) -> ResponseType:
        match self.request:
            case {'voice': _}:
                self.request.pop('voice')
                return await (await self.to_voice_api()).api()
            case _:
                return await self._fall_through_api()


class VoiceApi(GuildApi):
    def __init__(
        self, api: GuildApi, channel: discord.VoiceChannel | discord.StageChannel
    ) -> None:
        super().__init__(api, api.member)
        self.channel = channel
        self.mainservice = self.pi.mainservice

    async def to_main_api(self) -> 'MainApi':
        vc = await self.mainservice.raw_vc_for_member(self.member)
        main = await self.mainservice.mode(create=False, force_play=False).main_for_raw_vc(vc)
        return MainApi(self, vc, main)

    def _sub(self, request: dict) -> Self:
        return VoiceApi(super()._sub(request), self.channel)

    def _api_text(self) -> str:
        return 'voice api'

    async def _api(self) -> ResponseType:
        match self.request:
            case {'main': _}:
                self.request.pop('main')
                return await (await self.to_main_api()).api()
            case _:
                return await self._fall_through_api()


class MainApi(VoiceApi):
    def __init__(
        self, api: VoiceApi, vc: discord.VoiceClient, main: MainAudio
    ) -> None:
        super().__init__(api, api.channel)
        self.vc = vc
        self.main = main

    def _sub(self, request: dict) -> Self:
        return MainApi(super()._sub(request), self.vc, self.main)

    def _api_text(self) -> str:
        return 'main api'

    async def _api(self) -> ResponseType:
        match self.request:
            case {'type': 'volume'}:
                return self.main.volume
            case {'type': 'playing'}:
                return self.vc.is_playing()
            case {'type': 'paused'}:
                return self.vc.is_paused()
            case {'type': 'queueformat'}:
                return await self.main.queue.format()
            case {'type': 'queuejson'}:
                return await self.main.queue.pubjson(self.member, self.request.get('limit', 1000))
            case _:
                return await self._fall_through_api()


class OperatorApi(UserApi):
    def __init__(self, api: UserApi) -> None:
        super().__init__(api.session, api.request, api.user_id)

    def _guild_visible(self, guild: discord.Guild) -> bool:
        return True

    def _sub(self, request: dict) -> Self:
        return OperatorApi(super()._sub(request))

    def _api_text(self) -> str:
        return 'operator api'

    async def _api(self) -> ResponseType:
        match self.request:
            case {'target': str() as targetname}:
                return await InstrumentationApi(self, targetname).api()
            case {'type': 'resetmonitoring'}:
                return self.pi.mainservice.pmonitoring.reset()
            case {'type': 'guilds'}:
                guilds = []
                for guild in self.client.guilds:
                    if self._guild_visible(guild):
                        guilds.append(
                            {
                                'id': str(guild.id),
                                'member_count': guild.member_count,
                                'name': guild.name,
                            }
                        )
                return guilds
            case {'type': 'sleep', 'duration': (float() | int()) as duration, 'echo': _ as echo}:
                await asyncio.sleep(duration)
                return echo
            case {'type': 'pool'}:
                return self.pi.mainservice.pool_json()
            case _:
                return await self._fall_through_api()


class InstrumentationApi(OperatorApi):
    class UnknownTarget(Api.UnknownApi):
        def json(self) -> dict:
            return super().json() | {'unknowntarget': None}

    def __init__(self, api: OperatorApi, targetname: str) -> None:
        super().__init__(api)
        self.targets = self.pi.targets
        self.targetname = targetname
        target_tuple = self.targets.targets.get(targetname, None)
        if target_tuple is None:
            raise InstrumentationApi.UnknownTarget('unknown target', targetname)
        self.target, self.methodname = target_tuple.value

    def _sub(self, request: dict) -> Self:
        return InstrumentationApi(super()._sub(request), self.targetname)

    def _api_text(self) -> str:
        return 'instrumentation api'

    async def _api(self) -> ResponseType:
        match self.request:
            case {
                'type': str() as instrumentationname
            } if (
                instrumentation_factory := self.targets.instrumentations.get(instrumentationname)
            ) is not None:
                try:
                    instrumentation: Instrumentation = await self.pi.mainservice.pmonitoring.get(
                        self.targets.get_factory(
                            self.targetname,
                            self.target,
                            self.methodname,
                            instrumentationname,
                            instrumentation_factory.value,
                        )
                    )
                except KeyError as e:
                    raise InstrumentationApi.UnknownTarget(
                        'binding failed', self.targetname, instrumentationname, str(e)
                    ) from e
                if not isinstance(instrumentation, JsonLike):
                    raise TypeError
                return instrumentation.json()
            case _:
                return await self._fall_through_api()


class Count(Instrumentation, JsonLike):
    def __init__(self, target, methodname: str):
        super().__init__(target, methodname)
        self.count = 0

    def instrument(self, method, *args, **kwargs):
        self.count += 1
        return method(*args, **kwargs)

    def json(self) -> ResponseType:
        return self.count


class Concurrency(Instrumentation, JsonLike):
    def __init__(self, target, methodname: str):
        super().__init__(target, methodname)
        self.concurrency = 0

    async def instrument(self, method, *args, **kwargs):
        self.concurrency += 1
        try:
            return await method(*args, **kwargs)
        finally:
            self.concurrency -= 1

    def json(self) -> ResponseType:
        return self.concurrency
