__all__ = ('Monitoring', 'PersistentMonitoring')

import asyncio
from contextlib import AsyncExitStack, ExitStack
from typing import Any, Callable, Generic, TypeVar

from rainbowadn.instrument import Instrumentation

T = TypeVar('T', bound=Instrumentation, covariant=True)


class Provider(Generic[T]):
    def __init__(self, provider: Callable[[], T], /) -> None:
        self.provider = provider
        self.__count = 0
        self.__empty = asyncio.Event()
        self.__empty.set()
        self.__closed = False

    def __enter__(self) -> T:
        if self.__closed:
            raise RuntimeError('the provider is closed')
        if self.__count < 0:
            raise RuntimeError
        if self.__count == 0:
            self.__instrumentation = self.provider().__enter__()
            self.__empty.clear()
        self.__count += 1
        return self.__instrumentation

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.__count <= 0:
            raise RuntimeError
        self.__count -= 1
        if self.__count == 0:
            self.__empty.set()
            try:
                self.__instrumentation.__exit__(exc_type, exc_val, exc_tb)
            except:
                self.__closed = True
                raise
            finally:
                del self.__instrumentation

    async def stop(self) -> None:
        while self.__count:
            await self.__empty.wait()
        self.__closed = True


class ProviderManager(Generic[T]):
    def __init__(self, provider: Callable[[], T], /) -> None:
        self.provider = Provider(provider)

    async def __aenter__(self) -> Provider:
        return self.provider

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.provider.stop()


class Monitoring:
    async def get(self, provider: Callable[[], T]) -> Provider[T]:
        if provider not in self.__providers:
            self.__providers[provider] = asyncio.create_task(
                self.__es.enter_async_context(ProviderManager(provider))
            )
        return await self.__providers[provider]

    async def __aenter__(self) -> 'Monitoring':
        async with AsyncExitStack() as es:
            self.__providers: dict[
                Callable[[], Instrumentation],
                asyncio.Future[Provider]
            ] = {}
            self.__es = es.pop_all()
            return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        async with self.__es:
            del self.__es


class PersistentMonitoring:
    def __init__(self, monitoring: Monitoring) -> None:
        self.__monitoring = monitoring

    async def _get(self, provider: Callable[[], T]) -> T:
        return self.__es.enter_context(await self.__monitoring.get(provider))

    async def get(self, provider: Callable[[], T]) -> T:
        if provider not in self.__instrumentations:
            self.__instrumentations[provider] = asyncio.create_task(
                self._get(provider)
            )
        return await self.__instrumentations[provider]

    def __enter__(self) -> 'PersistentMonitoring':
        self.__instrumentations: dict[
            Callable[[], Instrumentation],
            asyncio.Future
        ] = {}
        self.__es = ExitStack()
        return self

    def reset(self) -> int:
        with self.__es:
            self.__es = ExitStack()
            length = len(self.__instrumentations)
            self.__instrumentations.clear()
            return length
        raise RuntimeError

    def __exit__(self, exc_type, exc_val, exc_tb):
        with self.__es:
            del self.__instrumentations
            del self.__es
